import re
from typing import Optional

from hexagon.runtime.update.cli.command import output_from_command_in_cli_project_path
from hexagon.support.storage import load_user_data, store_user_data


class CliGitConfig:
    def __init__(self, remote: str, head_branch: str) -> None:
        self.remote = remote
        self.head_branch = head_branch

    remote: str
    head_branch: str


CLI_GIT_CONFIG_STORAGE_KEY = "git-config"


def load_cli_git_config() -> Optional[CliGitConfig]:
    raw_data = load_user_data(CLI_GIT_CONFIG_STORAGE_KEY)
    try:
        write_dict = {}
        remote = raw_data["remote"] if raw_data and "remote" in raw_data else None
        if not remote:
            remotes = output_from_command_in_cli_project_path("git remote").splitlines()
            if len(remotes) == 1:
                remote = remotes[0]
            elif "origin" in remotes:
                remote = "origin"
            else:
                raise Exception(_("error.support.cli.git.failed_to_get_remote"))

            write_dict["remote"] = remote

        head_branch = (
            raw_data["head_branch"] if raw_data and "head_branch" in raw_data else None
        )
        if not head_branch:
            output = output_from_command_in_cli_project_path(
                f"git remote show {remote}"
            )
            head_branch = re.search(r"HEAD branch: (.+)", output).groups(0)[0]
            write_dict["head_branch"] = head_branch

        if write_dict:
            store_user_data(CLI_GIT_CONFIG_STORAGE_KEY, write_dict, append=True)

        return CliGitConfig(remote, head_branch)
    except Exception:
        return None



from typing import List

from hexagon.runtime.singletons import options
from hexagon.support.input.prompt.key_bindings import to_readable_name


def _fmt(msg: str, keys: List[str]) -> List[str]:
    return [msg.format(keys=" | ".join(keys))]


class HintsBuilder:
    def __init__(self):
        self.__top_hints = []
        self.__middle_hints = []
        self.__bottom_hints = []
        self.__has_enter_cancel_skip = False
        self.__has_autocomplete = False
        self.__has_select_toggles = False
        self.__has_vertical_movement = False
        self.__has_fuzzy_toggle = False
        self.__has_number_controls = False
        self.__has_floating_point = False
        self.__has_path_support = False

    def with_enter_cancel_skip(self) -> "HintsBuilder":
        if not self.__has_enter_cancel_skip:
            self.__bottom_hints += (
                _fmt(_("msg.support.prompt.hints.confirm"), ["ENTER"])
                + _fmt(_("msg.support.prompt.hints.cancel"), ["CTRL+C"])
                + _fmt(_("msg.support.prompt.hints.skip"), ["CTRL+Z"])
            )
            self.__has_enter_cancel_skip = True
        return self

    def with_autocomplete(self) -> "HintsBuilder":
        if not self.__has_autocomplete:
            self.__bottom_hints = (
                _fmt(_("msg.support.prompt.hints.autocomplete"), ["CTRL+SPACE"])
                + self.__bottom_hints
            )
            self.__has_autocomplete = True
        return self

    def with_select_toggles(self) -> "HintsBuilder":
        if not self.__has_select_toggles:
            self.__middle_hints += (
                _fmt(_("msg.support.prompt.hints.toggle"), ["SPACE"])
                + _fmt(_("msg.support.prompt.hints.toggle_move_down"), ["CTRL+I"])
                + _fmt(_("msg.support.prompt.hints.toggle_move_up"), ["SHIFT+TAB"])
                + _fmt(_("msg.support.prompt.hints.toggle_all"), ["ALT+R", "CTRL+R"])
                + _fmt(
                    _("msg.support.prompt.hints.toggle_all_true"), ["ALT+A", "CTRL+A"]
                )
            )
            self.__has_select_toggles = True
        return self

    def with_vertical_movement(self) -> "HintsBuilder":
        if not self.__has_vertical_movement:
            self.__top_hints += _fmt(
                _("msg.support.prompt.hints.move_down"), ["↓", "CTRL+N"]
            ) + _fmt(_("msg.support.prompt.hints.move_up"), ["↑", "CTRL+P"])
            self.__has_vertical_movement = True
        return self

    def with_fuzzy_toggle(self) -> "HintsBuilder":
        if not self.__has_fuzzy_toggle:
            self.__top_hints = (
                _fmt(_("msg.support.prompt.hints.toggle_fuzzy_search"), ["CTRL+F"])
                + self.__top_hints
            )
            self.__has_fuzzy_toggle = True
        return self

    def with_number_controls(self) -> "HintsBuilder":
        if not self.__has_number_controls:
            self.__top_hints += (
                _fmt(
                    _("msg.support.prompt.hints.number_decrement"),
                    ["↓", "CTRL+N"],
                )
                + _fmt(_("msg.support.prompt.hints.number_increment"), ["↑", "CTRL+P"])
                + _fmt(_("msg.support.prompt.hints.number_left"), ["←", "CTRL+B"])
                + _fmt(_("msg.support.prompt.hints.number_right"), ["→", "CTRL+F"])
            )
            self.__middle_hints = (
                _fmt(_("msg.support.prompt.hints.toggle_negative"), ["-"])
                + self.__middle_hints
            )
            self.__has_number_controls = True
        return self

    def with_floating_point(self) -> "HintsBuilder":
        if not self.__has_floating_point:
            self.__middle_hints += _fmt(
                _("msg.support.prompt.hints.number_dot"), ["."]
            ) + _fmt(
                _("msg.support.prompt.hints.alt_part_focus"), ["CTRL+I", "SHIFT+TAB"]
            )
            self.__has_floating_point = True
        return self

    def with_path_support(self) -> "HintsBuilder":
        if not self.__has_path_support:
            self.__top_hints += _fmt(
                _("msg.support.prompt.hints.create_path_at_location"),
                [to_readable_name(options.keymap.create_dir)],
            )
            self.__has_path_support = True
        return self

    def with_all(self) -> "HintsBuilder":
        return (
            self.with_enter_cancel_skip()
            .with_autocomplete()
            .with_select_toggles()
            .with_vertical_movement()
            .with_fuzzy_toggle()
        )

    def build(self) -> str:
        if not self.__top_hints and not self.__middle_hints and not self.__bottom_hints:
            raise ValueError(_("error.support.prompt.hints.no_hints_on_builder"))

        return "\n".join(
            [_("msg.support.prompt.hints.help")]
            + [
                " / ".join(line)
                for line in [self.__top_hints, self.__middle_hints, self.__bottom_hints]
                if line
            ]
        )



import re
import sys

from hexagon.runtime.dependencies import scan_and_install_dependencies
from hexagon.runtime.singletons import cli, configuration, options
from hexagon.runtime.update.cli.command import (
    execute_command_in_cli_project_path,
    output_from_command_in_cli_project_path,
)
from hexagon.runtime.update.cli.git import load_cli_git_config
from hexagon.runtime.update.shared import already_checked_for_updates
from hexagon.runtime.update.silent_fail import silent_fail
from hexagon.support.input.prompt import prompt
from hexagon.support.output.printer import log


@silent_fail
def check_for_cli_updates():
    if options.cli_update_disabled:
        return
    if already_checked_for_updates():
        return

    with log.status(_("msg.support.update.cli.checking_for_cli_updates")) as status:
        cli_git_config = load_cli_git_config()
        if not cli_git_config:
            return

        status.update(
            _("msg.support.update.cli.checking_for_cli_updates_on_branch").format(
                branch=_current_branch()
            )
        )

        if "is behind" in _get_updated_branch_status():
            log.info(
                _("msg.support.update.cli.new_version_available").format(
                    cli_name=cli.name
                )
            )
            if not prompt.confirm(
                _("action.support.update.cli.confirm_update"), default=True
            ):
                return

            _update_branch()
            scan_and_install_dependencies(configuration.custom_tools_path)

            log.info(_("msg.support.update.cli.updated"))
            log.finish()
            sys.exit(0)


def _current_branch():
    current_git_branch_status = output_from_command_in_cli_project_path("git status")
    return re.search(r"On branch (.+)", current_git_branch_status).groups(0)[0]


def _get_updated_branch_status():
    execute_command_in_cli_project_path("git remote update")
    branch_status = output_from_command_in_cli_project_path("git status -uno")
    return branch_status


def _update_branch():
    execute_command_in_cli_project_path("git pull", show_stdout=True)



import os
import subprocess
import sys

from hexagon.runtime.singletons import options
from hexagon.runtime.update import REPO_ORG, REPO_NAME
from hexagon.runtime.update import version
from hexagon.runtime.update.changelog.fetch import fetch_changelog
from hexagon.runtime.update.changelog.format import format_entries
from hexagon.runtime.update.changelog.parse import parse_changelog
from hexagon.runtime.update.shared import already_checked_for_updates
from hexagon.runtime.update.silent_fail import silent_fail
from hexagon.support.input.prompt import prompt
from hexagon.support.output.printer import log
from hexagon.support.storage import HEXAGON_STORAGE_APP

CHANGELOG_MAX_PRINT_ENTRIES = 10


@silent_fail
def check_for_hexagon_updates():
    if options.update_disabled:
        return
    if already_checked_for_updates(HEXAGON_STORAGE_APP):
        return

    with log.status(_("msg.support.update.hexagon.checking_new_versions")):
        current_version = version.local(
            override=os.getenv("HEXAGON_TEST_VERSION_OVERRIDE", None)
        )
        latest_version = version.latest()

        if current_version >= latest_version:
            return

        log.info(
            _("msg.support.update.hexagon.new_version_available").format(
                latest_version=latest_version
            )
        )

        if bool(os.getenv("HEXAGON_UPDATE_SHOW_CHANGELOG", "1")):
            with log.status(_("msg.support.update.hexagon.fetching_changelog")):
                changelog = parse_changelog(
                    current_version, fetch_changelog(REPO_ORG, REPO_NAME)
                )

            entries = format_entries(changelog)

            for entry in entries[:CHANGELOG_MAX_PRINT_ENTRIES]:
                log.info("  - " + entry.message)

            if len(entries) > CHANGELOG_MAX_PRINT_ENTRIES:
                log.info(_("msg.support.update.hexagon.and_much_more"))

        if not prompt.confirm(
            _("action.support.update.hexagon.confirm_update"), default=True
        ):
            return

        with log.status(_("msg.support.update.hexagon.updating")):
            subprocess.check_call(
                f"{sys.executable} -m pip --disable-pip-version-check install https://github.com/{REPO_ORG}/{REPO_NAME}/releases/download/v{latest_version}/hexagon-{latest_version}.tar.gz",
                shell=True,
                stdout=subprocess.DEVNULL,
            )

        log.info(_("msg.support.update.hexagon.updated"))
        log.finish()
        sys.exit(1)



import json
import sys
from urllib.request import Request, urlopen

from packaging.version import parse as parse_version

from hexagon.runtime.update import REPO_NAME, REPO_ORG
from hexagon.runtime.update.github import add_github_access_token


def local(override=None):
    return parse_version(override or _local_version())


def _local_version():
    if sys.version_info >= (3, 8):
        import importlib.metadata as importlib_metadata

        return importlib_metadata.version("hexagon")
    else:
        import pkg_resources

        return pkg_resources.require("hexagon")[0].version


def latest():
    latest_release_request = Request(
        f"https://api.github.com/repos/{REPO_ORG}/{REPO_NAME}/releases/latest"
    )
    add_github_access_token(latest_release_request)
    latest_github_release = json.load(urlopen(latest_release_request))
    return parse_version(latest_github_release["name"].replace("v", ""))



import datetime
from typing import Any, Dict, Optional

from pydantic import BaseSettings, ValidationError
from pydantic.types import DirectoryPath

from hexagon.runtime.yaml import YamlValidationError


def _local_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    from hexagon.support.storage import (
        HEXAGON_STORAGE_APP,
        HexagonStorageKeys,
        load_user_data,
    )

    return load_user_data(HexagonStorageKeys.options.value, HEXAGON_STORAGE_APP) or {}


def _save_settings_to_source(options):
    from hexagon.support.storage import (
        HEXAGON_STORAGE_APP,
        HexagonStorageKeys,
        store_user_data,
    )

    options_dict = options.dict()
    options_dict["update_time_between_checks"] = (
        options.update_time_between_checks.total_seconds()
    )
    store_user_data(
        HexagonStorageKeys.options.value, options_dict, app=HEXAGON_STORAGE_APP
    )


class KeymapOptions(BaseSettings):
    create_dir: str = "c-p"


class Options(BaseSettings):
    theme: Optional[str] = "default"
    update_time_between_checks: Optional[datetime.timedelta] = datetime.timedelta(
        days=1
    )
    send_telemetry: Optional[bool] = None
    disable_dependency_scan: Optional[bool] = False
    update_disabled: Optional[bool] = False
    cli_update_disabled: Optional[bool] = False
    config_storage_path: Optional[DirectoryPath] = None
    hints_disabled: Optional[bool] = False
    keymap: KeymapOptions = KeymapOptions()

    class Config:
        env_prefix = "HEXAGON_"

        @classmethod
        def customise_sources(cls, init_settings, env_settings, file_secret_settings):
            return (
                init_settings,
                env_settings,
                _local_settings_source,
                file_secret_settings,
            )


def get_options(init_settings: dict) -> Options:
    try:
        return Options(**init_settings)
    except ValidationError as errors:
        raise YamlValidationError(errors)


def update_options(opt: Options) -> Options:
    _save_settings_to_source(opt)
    return opt.copy()



from itertools import groupby
from typing import List

from hexagon.domain.cli import Cli
from hexagon.domain.env import Env
from hexagon.domain.tool import Tool
from hexagon.support.output.printer import log


def print_help(cli_config: Cli, tools: List[Tool], envs: List[Env]):
    """
    Print the command line help text based on the tools and envs in configuration yaml

    :param cli_config:
    :param tools:
    :param envs:
    :return:
    """
    if cli_config.name == "Hexagon":
        log.info("[bold]Hexagon", gap_end=1)
        log.info(_("msg.support.help.no_install"))
        log.info(_("msg.support.help.get_started"))
        return

    log.info(f"[bold]{cli_config.name}")
    log.info(
        f"usage: {cli_config.command} \[tool] \[env] \[\[positional-tool-arg] \[--optional-tool-arg=123] ...]",
        gap_start=1,
        gap_end=1,
    )

    log.info(_("msg.support.help.envs"))
    for i, env in enumerate(envs):
        log.info(
            f'  {env.name + (" (" + env.alias + ")" if env.alias else ""):<60}[dim]{env.long_name or ""}'
        )
        if env.description:
            log.info(
                f'  {"": <60}[dim]{env.description}',
                gap_end=__gap_if_last(envs, i),
            )
        else:
            log.gap()

    log.info(_("msg.support.help.tools"), gap_start=1)

    data = sorted(tools, key=lambda t: t.type, reverse=True)

    for gk, g in groupby(data, lambda t: t.type):
        log.info(f"[bold]{gk}:", gap_start=1)
        type_tools = list(g)

        for i, tool in enumerate(type_tools):
            log.info(
                f'  {tool.name + (" (" + tool.alias + ")" if tool.alias else ""):<60}[dim]{tool.long_name or ""}'
            )
            if tool.description:
                log.info(
                    f'  {"": <60}[dim]{tool.description}',
                    gap_end=__gap_if_last(type_tools, i),
                )


def __gap_if_last(envs, i):
    return 1 if i + 1 < len(envs) else 0



from enum import Enum

from hexagon.runtime.configuration import ConfigFile
from hexagon.support.input.args import ToolArgs, OptionalArg, Arg
from hexagon.support.input.types import FilePath
from hexagon.support.output.printer import log


class Target(Enum):
    STDOUT = "stdout"
    FILE = "file"


class Args(ToolArgs):
    target: OptionalArg[Target] = Arg(
        None, prompt_message="How do you want to output the schema?"
    )
    file_path: OptionalArg[FilePath] = Arg(
        None,
        prompt_message="Where do you want to save the schema?",
        allow_nonexistent=True,
    )


def main(tool, env, env_args, cli_args):
    if not cli_args.target.value:
        cli_args.target.prompt()

    if cli_args.target.value == Target.STDOUT:
        log.example(ConfigFile.schema_json(indent=2), syntax="json")
    else:
        if not cli_args.file_path.value:
            cli_args.file_path.prompt()
        with open(cli_args.file_path.value, "w") as schema_file:
            schema_file.write(ConfigFile.schema_json(indent=2))

        log.result(f"[green]Schema saved to [b]{cli_args.file_path.value}[/green]")



import re
from io import StringIO
from typing import List, Optional

from markdown import Markdown
from packaging.version import parse as parse_version, Version

from hexagon.runtime.update.changelog import ChangelogVersionEntry, ChangelogEntry
from hexagon.runtime.update.changelog.fetch import ChangelogFile

CHANGELOG_MAX_EMPTY_LINES = 10


def parse_changelog(
    current_hexagon_version: Version, changelog_file: ChangelogFile
) -> List[ChangelogVersionEntry]:
    entries = []
    current_version: Optional[ChangelogVersionEntry] = None
    current_entry_type: str = ""
    consecutive_empty_lines_count = 0
    while True:
        line = changelog_file.line()

        version_match = re.search("^## v(\\d+\\.\\d+\\.\\d+)", line)
        if version_match:
            if current_version:
                entries.append(current_version)
            current_version = ChangelogVersionEntry(version_match.groups(0)[0])
            if current_hexagon_version == parse_version(current_version.version):
                entries.append(current_version)
                break

        entry_type_match = re.search("^### (\\w+)$", line)
        if entry_type_match:
            current_entry_type = entry_type_match.groups(0)[0]

        match = re.search("^\\* ([^(]+)", line)
        if match:
            current_version.entries.append(
                ChangelogEntry(current_entry_type, _unmark(match.groups(0)[0]))
            )

        if not line:
            if consecutive_empty_lines_count > CHANGELOG_MAX_EMPTY_LINES:
                if current_version:
                    entries.append(current_version)
                break
            else:
                consecutive_empty_lines_count += 1
        else:
            consecutive_empty_lines_count = 0

    return entries


def _unmark_element(element, stream=None):
    if stream is None:
        stream = StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        _unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


Markdown.output_formats["plain"] = _unmark_element
# patching Markdown
# noinspection PyTypeChecker
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False


def _unmark(text):
    return __md.convert(text)



import os
from typing import Any, Type

from pydantic import ValidationError
from ruamel import yaml
from ruamel.yaml import YAML

from hexagon.domain.hexagon_error import HexagonError
from hexagon.support.output.printer import Logger


def read_file(path: str) -> Any:
    try:
        return YAML().load(open(path, "r"))
    except FileNotFoundError:
        return None


def write_file(path: str, content: str):
    with open(path, "w") as f:
        YAML().dump(content, f)
    return True


def load_model(model: Type, content: dict, path: str):
    try:
        return model(**content)
    except ValidationError as error:
        raise YamlValidationError(error, content, path)


class YamlValidationError(HexagonError):
    def __init__(self, error: ValidationError, yaml_content=None, yaml_path=None):
        self.yaml_content = yaml_content
        self.yaml_path = yaml_path
        self.errors = error.errors()
        self.actual_yaml_file = open(yaml_path, "r").read() if yaml_path else None
        super().__init__(self._error_printer)

    def _error_printer(self, logger: Logger):
        logger.error(
            _("error.support.yaml.invalid_yaml").format(
                errors_length=len(self.errors), yaml_path=self.yaml_path
            )
        )
        for err in self.errors:
            logger.error(
                os.linesep  # this \n can not go in the .po file because it breaks msgmt
                + _("error.support.yaml.error_at").format(
                    loc=".".join(map(lambda i: str(i), err["loc"])), message=err["msg"]
                )
            )
            if self.yaml_content and yaml:
                (start, line_number, end) = self.__lines_of_error(
                    err, self.yaml_content
                )
                logger.example(
                    "\n".join(self.actual_yaml_file.splitlines()[start:end]),
                    syntax="yaml",
                    show_line_numbers=True,
                    start_line=start + 1,
                    decorator_start=False,
                    decorator_end=False,
                )

    @classmethod
    def __lines_of_error(cls, err, ruamel_yaml):
        line_number = cls.__yaml_line_number(ruamel_yaml, err["loc"])
        return (
            max(0, line_number - 3),
            line_number,
            line_number + 2 if line_number != 0 else line_number,
        )

    @classmethod
    def __yaml_line_number(cls, yml, loc: list, line_count_hack: int = 0):
        """
        Get line number of location by accessing YAML metadata

        :param yml: a dict representing de YAML, usually created with YAML().load(file)
        :param loc: a list of the keys in the YAML
        :param line_count_hack: a counter to indicate nested level,
        for some reason ruamel .lc returns one less line_number for each level
        :return: the line number of loc
        """
        if len(loc) == 1:
            try:
                return (
                    line_count_hack + yml[loc[0]].lc.line
                    if yml[loc[0]]
                    else yml.lc.data[loc[0]][2]
                )
            except (LookupError, TypeError, AttributeError):
                return (
                    line_count_hack - 1 if line_count_hack > 0 else line_count_hack
                ) + yml.lc.line
        else:
            return cls.__yaml_line_number(
                yml[loc[0]], loc[1:], line_count_hack=line_count_hack + 1
            )



import abc
from copy import copy
from typing import Generic, Union, Callable, TypeVar

from pydantic import ValidationError
from pydantic.fields import ModelField

ARGUMENT_KEY_PREFIX = "-"

T = TypeVar("T")


class HexagonArg(Generic[T]):
    __model__ = None
    __field__ = None

    def __init__(self, value: T):
        self.__value__ = value

    @classmethod
    def __get_validators__(cls):
        yield cls.arg_validator(cls)

    @property
    def value(self):
        if self.__model__:
            self.__model__.trace(self.__field__)
        return self.__value__

    def prompt(
        self,
        skip_trace: Union[bool, Callable] = False,
        **kwargs,
    ):
        """
        Prompt the user for a value for this argument.

        :param skip_trace: true if this prompt should not be traced by hexagon
        :param kwargs: any extra argument supported by hexagon and InquirePy
        :return:
        """
        if not self.__model__:
            raise ValueError(
                "Cannot prompt for a value when model reference is not initialized. "
                "Probably _init_refs was not called."
            )
        return self.__model__.prompt(self.__field__, skip_trace, **kwargs)

    def _init_refs(self, model, field):
        self.__field__ = field
        self.__model__ = model

    @staticmethod
    @abc.abstractmethod
    def cli_repr(field: ModelField):
        return

    def __str__(self, **kwargs):
        return str(self.__value__)

    @staticmethod
    def arg_validator(cls):
        def validate(v, field: ModelField):
            if not isinstance(v, cls):
                v = cls(v)
            if not field.sub_fields:
                # Generic parameters were not provided, ie: `name = PositionalArg`,
                # so we don't try to validate them and just return the value as is
                return v
            value = field.sub_fields[0]

            value.field_info.extra = copy(field.field_info.extra)
            valid_value, error = value.validate(v.__value__, {}, loc="")
            if error:
                raise ValidationError([error], model=cls)
            return cls(valid_value)

        return validate


class PositionalArg(HexagonArg[T]):
    @staticmethod
    def cli_repr(field: ModelField):
        return field.name, None


class OptionalArg(HexagonArg[T]):
    @staticmethod
    def cli_repr(field: ModelField):
        def initials_from(name: str):
            return "".join([w[0] for w in name.split("_")])

        return (
            name_key(field.name.replace("_", "-")),
            alias_key(
                field.alias.replace("_", "-")
                if field.alias and field.alias != field.name
                else initials_from(field.name)
            ),
        )


def name_key(name: str):
    return f"{ARGUMENT_KEY_PREFIX * 2}{name}"


def alias_key(alias: str):
    return f"{ARGUMENT_KEY_PREFIX}{alias}"



import importlib
import os
import sys
from typing import List


def collect_plugins(project_path: str, plugins_list: List[str]):
    _import_all_in_folder(os.path.dirname(__file__))

    for plugin in plugins_list:
        plugins_dir_path = os.path.abspath(os.path.join(project_path, plugin))

        if os.path.isdir(plugins_dir_path):
            _import_all_in_folder(plugins_dir_path)
        elif os.path.isfile(plugins_dir_path):
            _import_plugin(plugin)


def _import_all_in_folder(path: str):
    sys.path.append(os.path.abspath(path))
    for module in os.listdir(path):
        _import_plugin(module)


def _import_plugin(module):
    name = os.path.splitext(module)[0]
    name = name.replace(os.path.sep, ".")

    if name != "__init__" and name != "__pycache__":
        imported = importlib.import_module(name)
        imported.main()



import re
from copy import copy
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, Literal, Callable, List

from InquirerPy import inquirer
from InquirerPy.base import Choice
from InquirerPy.utils import run_in_terminal
from prompt_toolkit.buffer import ValidationState
from prompt_toolkit.document import Document
from prompt_toolkit.validation import ValidationError, Validator
from pydantic import ValidationError as PydanticValidationError, DirectoryPath
from pydantic.fields import ModelField, Validator as PydanticValidator

from hexagon.domain.hexagon_error import ListHexagonError
from hexagon.runtime.singletons import options
from hexagon.support.input.args import HexagonArg
from hexagon.support.input.prompt.for_all_methods import for_all_methods
from hexagon.support.input.prompt.hints import HintsBuilder
from hexagon.support.input.types import path_validator
from hexagon.support.output.printer import log
from hexagon.utils.typing import field_info


class HexagonArgumentSetupError(ListHexagonError):
    def __init__(self, argument: str, prop: str):
        super().__init__(
            [
                _("error.support.input.prompt.prompt.invalid_argument_setup").format(
                    argument=argument, property=prop
                ),
            ]
        )


class PromptValidator(Validator):
    def __init__(self, validators, cls):
        self.validators = validators
        self.cls = cls

    def validate(self, document: Document) -> None:
        try:
            for validator in self.validators.values():
                validator.func(self.cls, document.text)
        except PydanticValidationError as e:
            raise ValidationError(
                message=" / ".join([x["msg"] for x in e.errors()]),
                cursor_position=len(document.text),
            )
        except ValueError as e:
            raise ValidationError(message=e.args[0], cursor_position=len(document.text))


def default_validator(model_field: ModelField, mapper=lambda x: x):
    def func(cls, value):
        value, error = model_field.validate(mapper(value), {}, loc="")
        if error:
            raise PydanticValidationError([error], model=cls)
        return value

    return func


def list_mapper(v):
    return v.strip().split("\n")


def set_default(invocation_extras, model_field: ModelField):
    if model_field.field_info.extra.get("prompt_default"):
        return {"default": model_field.field_info.extra["prompt_default"]}
    if "default" in invocation_extras:
        return {"default": invocation_extras["default"]}
    elif model_field.default:
        default = (
            model_field.default.value
            if isinstance(model_field.default, HexagonArg)
            else model_field.default
        )
        if default is not None:
            return {"default": default}

    return {}


class InquiryType(Enum):
    STRING = auto()
    STRING_SEARCHABLE = auto()
    STRING_LIST = auto()
    STRING_LIST_SEARCHABLE = auto()
    ENUM = auto()
    ENUM_SEARCHABLE = auto()
    ENUM_LIST = auto()
    ENUM_LIST_SEARCHABLE = auto()
    PATH = auto()
    PATH_SEARCHABLE = auto()
    INT = auto()
    FLOAT = auto()
    SECRET = auto()
    BOOLEAN = auto()


def _determine_expected_inquiry(
    iterable, of_enum, type_, declaration_extras, invocation_extras
) -> (InquiryType, dict):
    """
    TODO: move this to a separate module and unit test it
    """
    extras = {**declaration_extras, **invocation_extras}
    searchable = extras.get("searchable", False)

    query = InquiryType.STRING if not searchable else InquiryType.STRING_SEARCHABLE
    if iterable and of_enum:
        query = (
            InquiryType.ENUM_LIST
            if not searchable
            else InquiryType.ENUM_LIST_SEARCHABLE
        )
    elif iterable:
        query = (
            InquiryType.STRING_LIST
            if not searchable
            else InquiryType.STRING_LIST_SEARCHABLE
        )
    elif of_enum:
        query = InquiryType.ENUM if not searchable else InquiryType.ENUM_SEARCHABLE
    elif issubclass(type_, Path):
        query = InquiryType.PATH if not searchable else InquiryType.PATH_SEARCHABLE
    elif issubclass(type_, bool):
        query = InquiryType.BOOLEAN
    elif issubclass(type_, int):
        query = InquiryType.INT
    elif issubclass(type_, float):
        query = InquiryType.FLOAT
    elif extras.get("secret", False):
        query = InquiryType.SECRET

    return query, extras


@for_all_methods(log.status_aware, exclude=["query_field"])
class Prompt:
    def query_field(self, model_field: ModelField, model_class, **kwargs):
        # TODO: this method is a mess, refactor it
        declaration_extras = copy(model_field.field_info.extra)
        invocation_extras = copy(kwargs)
        inquiry_args = {
            "message": declaration_extras.get("prompt_message", None)
            or f"Enter {model_field.name}:",
            "instruction": declaration_extras.get("prompt_instruction", None),
        }
        inquiry_args.update(set_default(invocation_extras, model_field))

        field_type, iterable, of_enum = field_info(model_field)

        inquiry_type, extras = _determine_expected_inquiry(
            iterable, of_enum, field_type, declaration_extras, invocation_extras
        )

        setups = {
            InquiryType.ENUM_LIST: setup_enum_list,
            InquiryType.ENUM_LIST_SEARCHABLE: setup_enum_list,
            InquiryType.STRING_LIST: setup_string_list,
            InquiryType.STRING_LIST_SEARCHABLE: setup_searchable_list,
            InquiryType.ENUM: setup_enum,
            InquiryType.ENUM_SEARCHABLE: setup_enum,
            InquiryType.STRING_SEARCHABLE: setup_searchable_list,
            InquiryType.PATH: setup_path,
            InquiryType.PATH_SEARCHABLE: setup_path,
            InquiryType.INT: setup_int,
            InquiryType.FLOAT: setup_float,
            InquiryType.SECRET: setup_secret,
            InquiryType.BOOLEAN: setup_bool,
        }

        inq, mapper = setups.get(inquiry_type, setup_string)(
            self,
            inquiry_args=inquiry_args,
            inquiry_type=inquiry_type,
            extras=extras,
            field_type=field_type,
        )

        if model_field.sub_fields:
            validators_ = {
                "default": PydanticValidator(
                    default_validator(model_field, mapper=mapper), check_fields=True
                )
            }
            if model_field.class_validators:
                validators_.update(
                    {
                        **model_field.class_validators,
                    }
                )
            inquiry_args["validate"] = PromptValidator(
                validators_,
                model_class,
            )

        # FIXME: this is working by chance, it's not the best way to do it
        if "searchable" in invocation_extras:
            del invocation_extras["searchable"]

        inquiry_args.update(**invocation_extras)
        try:
            return inq(**inquiry_args)
        except TypeError as e:
            if "__init__() got an unexpected keyword argument" in e.args[0]:
                prop = re.search(".*\s('\w+')$", e.args[0]).group(1).strip()
                err = HexagonArgumentSetupError(model_field.name, prop)
            else:
                raise e

        if err:
            raise err

    @staticmethod
    def text(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = (
                HintsBuilder().with_enter_cancel_skip().with_autocomplete().build()
            )
        return inquirer.text(**kwargs).execute()

    @staticmethod
    def select(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = (
                HintsBuilder().with_enter_cancel_skip().with_vertical_movement().build()
            )
        return inquirer.select(**kwargs).execute()

    @staticmethod
    def checkbox(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = (
                HintsBuilder()
                .with_enter_cancel_skip()
                .with_vertical_movement()
                .with_select_toggles()
                .build()
            )
        return inquirer.checkbox(**kwargs).execute()

    @staticmethod
    def confirm(*args, **kwargs):
        if not options.hints_disabled:
            # not adding confirm_letter and reject_letter as hints because it seems redundant
            kwargs["long_instruction"] = HintsBuilder().with_enter_cancel_skip().build()
        if "validate" in kwargs:
            del kwargs["validate"]  # validate is not supported by confirm
        return inquirer.confirm(*args, **kwargs).execute()

    @staticmethod
    def fuzzy(**kwargs):
        kwargs["keybindings"] = {
            "toggle-exact": [
                {"key": "c-f"}
            ]  # toggle string matching algorithm between fuzzy or exact
        }
        if not options.hints_disabled:
            builder = (
                HintsBuilder()
                .with_enter_cancel_skip()
                .with_vertical_movement()
                .with_fuzzy_toggle()
            )
            if "multiselect" in kwargs:
                builder.with_select_toggles()
            kwargs["long_instruction"] = builder.build()
        return inquirer.fuzzy(border=log.use_borders(), **kwargs).execute()

    @staticmethod
    def path(**kwargs):
        is_dir = kwargs.get("only_directories", False)
        if not options.hints_disabled:
            hints = HintsBuilder().with_enter_cancel_skip().with_autocomplete()
            if is_dir:
                hints.with_path_support()
            kwargs["long_instruction"] = hints.build()
        filepath_prompt = inquirer.filepath(**kwargs)

        if is_dir:
            # noinspection PyProtectedMember
            @filepath_prompt.register_kb(options.keymap.create_dir)
            def _create_dir(__):
                path = path_validator(filepath_prompt._session.default_buffer.text)
                if not path.exists():
                    try:
                        path.mkdir(parents=True)
                        run_in_terminal(
                            lambda: log.info(
                                _("msg.support.prompt.prompt.directory_created").format(
                                    path=path.absolute()
                                )
                            )
                        )
                        filepath_prompt._session.default_buffer.validation_state = (
                            ValidationState.INVALID
                        )
                        filepath_prompt._session.default_buffer.validation_error = None
                    except Exception as e:
                        filepath_prompt._set_error(str(e))

        return filepath_prompt.execute()

    @staticmethod
    def number(**kwargs):
        if not options.hints_disabled:
            builder = HintsBuilder().with_enter_cancel_skip().with_number_controls()
            if "float_allowed" in kwargs:
                builder.with_floating_point()
            kwargs["long_instruction"] = builder.build()
        return inquirer.number(**kwargs).execute()

    @staticmethod
    def secret(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = HintsBuilder().with_enter_cancel_skip().build()
        return inquirer.secret(**kwargs).execute()


def setup_enum_list(
    self: Prompt,
    inquiry_type: Literal[InquiryType.ENUM_LIST, InquiryType.ENUM_LIST_SEARCHABLE],
    inquiry_args: Dict[str, Any],
    field_type: Any,
    **_,
) -> (Callable, Callable):
    inquiry_args["choices"] = [
        Choice(
            name=x.name,
            value=x,
            enabled=("default" in inquiry_args and x in inquiry_args["default"]),
        )
        for x in field_type.__args__[0]
    ]
    inquiry_args["default"] = ""
    if inquiry_type == InquiryType.ENUM_LIST_SEARCHABLE:
        inquiry_args["multiselect"] = True
        return self.fuzzy, lambda x: x
    else:
        return self.checkbox, lambda x: x


def setup_string_list(
    self: Prompt, inquiry_args: Dict[str, Any], **_
) -> (Callable, Callable):
    mapper = list_mapper
    inquiry_args["filter"] = mapper
    inquiry_args["instruction"] = (
        inquiry_args["instruction"]
        or "(each line represents a value) ESC + Enter to finish input"
    )
    inquiry_args["multiline"] = True
    return self.text, mapper


def setup_enum(
    self: Prompt,
    inquiry_type: Literal[InquiryType.ENUM, InquiryType.ENUM_SEARCHABLE],
    inquiry_args: Dict[str, Any],
    field_type: Any,
    **_,
) -> (Callable, Callable):
    inquiry_args["choices"] = [{"name": x.name, "value": x} for x in field_type]
    if inquiry_type == InquiryType.ENUM_SEARCHABLE:
        inquiry_args["default"] = (
            inquiry_args["default"].value if "default" in inquiry_args else None
        )
        return self.fuzzy, lambda x: x
    else:
        return self.select, lambda x: x


def setup_searchable_list(
    self: Prompt,
    inquiry_type: Literal[
        InquiryType.STRING_SEARCHABLE, InquiryType.STRING_LIST_SEARCHABLE
    ],
    inquiry_args: Dict[str, Any],
    extras: Dict[str, Any],
    **_,
) -> (Callable, Callable):
    inquiry_args["choices"] = extras["choices"]
    if inquiry_type == InquiryType.STRING_LIST_SEARCHABLE:
        inquiry_args["multiselect"] = True
    return self.fuzzy, lambda x: x


def setup_path(
    self: Prompt,
    inquiry_type: Literal[InquiryType.PATH, InquiryType.PATH_SEARCHABLE],
    inquiry_args: Dict[str, Any],
    extras: Dict[str, Any],
    field_type: Any,
) -> (Callable, Callable):
    if inquiry_type == InquiryType.PATH_SEARCHABLE:
        choices: List[Dict[str, Any] or Any] = []
        if "glob_extra_choices" in extras:
            extra_choices = extras["glob_extra_choices"]
            for choice in extra_choices:
                choices.append(
                    choice
                    if isinstance(choice, dict)
                    else {
                        "name": choice,
                        "value": choice,
                    }
                )

        if "glob" in extras:
            choices = choices + [
                f
                for f in sorted(
                    Path(extras.get("cwd", ".")).rglob(extras["glob"]),
                    key=lambda x: x.name,
                )
            ]
        elif "choices" in extras:
            choices = choices + extras["choices"]
        else:
            raise ValueError("searchable path must have either glob or choices defined")
        inquiry_args["choices"] = choices
        return self.fuzzy, lambda x: x
    else:
        inquiry_args["only_directories"] = (
            inquiry_args["only_directories"]
            if "only_directories" in inquiry_args
            else issubclass(field_type, DirectoryPath)
        )
        return self.path, lambda x: x


def setup_int(self: Prompt, **_) -> (Callable, Callable):
    return self.number, lambda x: x


def setup_float(
    self: Prompt, inquiry_args: Dict[str, Any], **_
) -> (Callable, Callable):
    inquiry_args["float_allowed"] = True
    return self.number, lambda x: x


def setup_secret(self: Prompt, **_) -> (Callable, Callable):
    return self.secret, lambda x: x


def setup_string(self: Prompt, **_) -> (Callable, Callable):
    return self.text, lambda x: x


def setup_bool(self: Prompt, **_) -> (Callable, Callable):
    return self.confirm, lambda x: x



from typing import Optional, List

from pydantic import BaseModel


class Cli(BaseModel):
    name: str
    command: str
    custom_tools_dir: Optional[str] = None
    plugins: List[str] = []
    options: Optional[dict] = None



from __future__ import annotations

from enum import Enum
from typing import Callable, List, Optional, Dict, Any, Union

from pydantic import BaseModel


class ToolType(str, Enum):
    misc = "misc"
    web = "web"
    shell = "shell"
    hexagon = "hexagon"
    group = "group"
    function = "function"


class Tool(BaseModel):
    name: str
    type: ToolType = ToolType.misc
    icon: Optional[str] = None
    alias: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    envs: Optional[Dict[str, Any]] = None
    traced: Optional[bool] = True

    class Config:
        use_enum_values = True


class ActionTool(Tool):
    action: Union[str, List[str]]

    @property
    def executable_str(self):
        if isinstance(self.action, list):
            return "\n".join(self.action)
        cmd = str(self.action)
        return cmd[:-1] if cmd.endswith("\n") else cmd


class FunctionTool(Tool):
    function: Callable


class GroupTool(Tool):
    tools: Union[str, List[Union[ActionTool, GroupTool]]]


GroupTool.update_forward_refs()


class ToolGroupConfigFile(BaseModel):
    tools: List[Union[ActionTool, GroupTool, FunctionTool]]



def merge_dictionaries_deep(a: dict, b: dict, path=None):
    """merges b into a"""
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dictionaries_deep(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a



from typing import Optional, Dict, Union, List

from pydantic import BaseModel

from hexagon.support.input.args.hexagon_args import PositionalArg


class CliArgs(BaseModel):
    show_help: bool = False
    tool: PositionalArg[Optional[str]] = None
    env: PositionalArg[Optional[str]] = None

    extra_args: Optional[Dict[str, Union[list, bool, int, str]]] = None
    raw_extra_args: Optional[List[str]] = None
    total_args: int

    def as_list(self):
        return [str(x) for x in [self.tool, self.env] if x] + (
            self.raw_extra_args if self.raw_extra_args else []
        )

    def as_str(self):
        return " ".join(self.as_list())

    def count(self):
        return self.total_args

    @staticmethod
    def key_value_arg(key, arg):
        return f"{key}={arg}"

    @property
    def format_friendly_extra_args(self):
        return (
            {
                "positional": sorted(
                    [v for k, v in self.extra_args.items() if k.isdigit()]
                ),
                "optional": {
                    k: v for k, v in self.extra_args.items() if not k.isdigit()
                },
            }
            if self.extra_args
            else {}
        )



from typing import Optional, Union, Any, Callable

from pydantic import (
    Field as PydanticField,
)
from hexagon.support.input.args.hexagon_args import (
    ARGUMENT_KEY_PREFIX,
    HexagonArg,
    PositionalArg,
    OptionalArg,
)
from hexagon.support.input.args.tool_args import ToolArgs
from hexagon.support.input.args.cli_args import CliArgs

ARGUMENT_KEY_PREFIX = ARGUMENT_KEY_PREFIX
HexagonArg = HexagonArg
PositionalArg = PositionalArg
OptionalArg = OptionalArg
ToolArgs = ToolArgs
CliArgs = CliArgs


# noinspection PyPep8Naming
def Arg(
    default: Any = None,
    *,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    prompt_default: Optional[Union[Any, Callable[[Any], Any]]] = None,
    prompt_message: Optional[Union[str, Callable[[Any], str]]] = None,
    prompt_instruction: Optional[str] = None,
    searchable: bool = False,
    **kwargs,
):
    """
    Used to provide extra information about an argument, either for the model schema or complex validation.
    Some arguments apply only to number fields (``int``, ``float``, ``Decimal``) and some apply only to ``str``.

    TODO: add support for `validators` kwarg
    """
    return PydanticField(
        default,
        alias=alias,
        title=title,
        description=description,
        prompt_default=prompt_default,
        prompt_message=prompt_message,
        prompt_instruction=prompt_instruction,
        searchable=searchable,
        **kwargs,
    )



def silent_fail(func):
    """
    Decorator to call a function and catch any exception that might be raised.

    :return: a wrapper for the decorated function
    """

    def wrapper():
        # noinspection PyBroadException
        try:
            return func()
        except Exception:
            return

    return wrapper



from copy import copy
from pathlib import Path
from typing import Any

from pydantic import (
    FilePath as PydanticFilePath,
    DirectoryPath as PydanticDirectoryPath,
    errors,
)
from pydantic.errors import _PathValueError


class FileNotExistsError(_PathValueError):
    code = "path.not_exists"
    msg_template = 'file at path "{path}" does not exist.'


class DirectoryNotExistsError(_PathValueError):
    code = "path.not_exists"
    msg_template = 'directory at path "{path}" does not exist.'


class FilePath(PydanticFilePath):
    @classmethod
    def __get_validators__(cls, *args, **kwargs):
        yield path_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: Path, field) -> Path:
        declaration_extras = copy(field.field_info.extra)

        extra_choices = declaration_extras.get("glob_extra_choices")
        if extra_choices:
            for extra_choice in extra_choices:
                if (
                    value.name == extra_choice
                    if isinstance(extra_choice, str)
                    else extra_choice["value"]
                ):
                    return value

        if not declaration_extras.get("allow_nonexistent", False):
            if not value.exists():
                raise FileNotExistsError(path=value)

            if not value.is_file():
                raise errors.PathNotAFileError(path=value)

        return value


class DirectoryPath(PydanticDirectoryPath):
    @classmethod
    def __get_validators__(cls):
        yield path_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: Path, field) -> Path:
        declaration_extras = copy(field.field_info.extra)

        extra_choices = declaration_extras.get("glob_extra_choices")
        if extra_choices:
            for extra_choice in extra_choices:
                if (
                    value.name == extra_choice
                    if isinstance(extra_choice, str)
                    else extra_choice["value"]
                ):
                    return value

        if not declaration_extras.get("allow_nonexistent", False):
            if not value.exists():
                raise DirectoryNotExistsError(path=value)

            if not value.is_dir():
                raise errors.PathNotADirectoryError(path=value)

        return value


def path_validator(v: Any) -> Path:
    if isinstance(v, Path):
        return v

    try:
        return Path(v)
    except TypeError:
        raise errors.PathError()



from enum import Enum
from typing import Any, Union

from pydantic.fields import ModelField


def should_support_multiple_args(field: Union[ModelField, Any]):
    type_ = _field_type(field) if isinstance(field, ModelField) else field
    if hasattr(type_, "__origin__"):
        type_ = type_.__origin__
    return type_ in [list, tuple, set]


def field_info(field: ModelField):
    type_ = _field_type(field)
    iterable = should_support_multiple_args(field)
    if iterable and hasattr(type_, "__args__") and len(type_.__args__):
        of_enum = _is_enumerable(type_.__args__[0])
    else:
        of_enum = _is_enumerable(type_)

    return type_, iterable, of_enum


def _field_type(field: ModelField):
    return field.sub_fields[0].outer_type_


def _is_enumerable(type_):
    try:
        res = issubclass(type_, Enum)
    except TypeError:
        res = False
    return res



import os
import sys
from enum import Enum
from pathlib import Path
from shutil import rmtree
from typing import Any, Dict, List

from ruamel.yaml import YAML

from hexagon.support.storage.merge_dictionaries import merge_dictionaries_deep

HEXAGON_STORAGE_APP = "hexagon"


class HexagonStorageKeys(Enum):
    options = "options"
    last_command = "last-command"
    last_update_check = "last-update-check"
    cli_install_path = "cli-install-path"


class StorageValueType(Enum):
    text = "text"
    text_multiline = "text-multiline"
    dictionary = "dictionary"


class StoragePurpose(Enum):
    config = "config"
    data = "local/share"


_extension_by_value_type = {
    StorageValueType.text: ".txt",
    StorageValueType.text_multiline: ".txt",
    StorageValueType.dictionary: ".yaml",
}

_config_storage_path = None
_data_storage_path = None

InputDataType = str or List[str] or Dict[str, Any]


def __storage_path_by_os(purpose: StoragePurpose):
    return {
        "linux": os.path.expanduser(f"~/.{purpose.value}/hexagon"),
        "darwin": os.path.expanduser(f"~/.{purpose.value}/hexagon"),
        "cygwin": os.path.expanduser(f"~/.{purpose.value}/hexagon"),
        "win32": os.path.expanduser("~/hexagon"),
    }


def _get_storage_dir_path():
    global _config_storage_path
    if _config_storage_path:
        return _config_storage_path

    _config_storage_path = os.getenv(
        "HEXAGON_STORAGE_PATH",
        __storage_path_by_os(StoragePurpose.config)[sys.platform],
    )
    Path(_config_storage_path).mkdir(exist_ok=True)

    return _config_storage_path


def _resolve_storage_path(app: str, key: str, base_dir=None):
    base_dir = base_dir if base_dir else _get_storage_dir_path()
    key_split = key.split(".")
    # noinspection PyRedundantParentheses
    return (os.path.join(base_dir, app, *key_split[:-1]), *key_split[-1:])


def _storage_value_type_by_data_type(data: InputDataType):
    if isinstance(data, str):
        return StorageValueType.text, data
    elif isinstance(data, list):
        return StorageValueType.text_multiline, [str(s) for s in data]
    elif isinstance(data, dict):
        return StorageValueType.dictionary, data
    else:
        raise Exception(
            f"Type {type(data)} cannot be stored: supported types are str, List[str] or Dict"
        )


def _storage_value_type_by_file_path(file_path: str):
    if os.path.exists(file_path + ".txt"):
        return StorageValueType.text
    elif os.path.exists(file_path + ".yaml"):
        return StorageValueType.dictionary
    else:
        return None


def _storage_file(dir_path: str, file_name: str):
    base_file_path = os.path.join(dir_path, file_name)
    value_type = _storage_value_type_by_file_path(base_file_path)
    if not value_type:
        return None, None
    file_path = base_file_path + _extension_by_value_type[value_type]
    return file_path, value_type


def _get_app(app: str = None):
    from hexagon.runtime.singletons import cli, configuration

    return (
        app
        if app
        else (
            cli.name.lower()
            if cli and configuration.has_config
            else HEXAGON_STORAGE_APP
        )
    )


def store_user_data(key: str, data: InputDataType, append=False, app: str = None):
    app = _get_app(app)
    value_type, sanitized_data = _storage_value_type_by_data_type(data)
    extension = _extension_by_value_type[value_type]
    dir_path, file_name = _resolve_storage_path(app, key)
    file_name += extension
    file_path = os.path.join(dir_path, file_name)

    Path(dir_path).mkdir(exist_ok=True, parents=True)

    if (
        value_type == StorageValueType.text_multiline
        or value_type == StorageValueType.text
    ):
        with open(file_path, "a" if append else "w") as file:
            if isinstance(sanitized_data, list):
                file.writelines(line + "\n" for line in sanitized_data)
            else:
                file.write(sanitized_data)

    elif value_type == StorageValueType.dictionary:
        previous = None
        if append and os.path.exists(file_path):
            with open(file_path, "r") as file:
                previous = YAML().load(file)

        to_write = (
            merge_dictionaries_deep(previous, sanitized_data)
            if previous
            else sanitized_data
        )

        with open(file_path, "w") as file:
            YAML().dump(to_write, file)


def load_user_data(key: str, app: str = None):
    app = _get_app(app)
    file_path, value_type = _storage_file(*_resolve_storage_path(app, key))

    if not value_type:
        return None

    if not os.path.isfile(file_path):
        return None

    with open(file_path, "r") as file:
        if (
            value_type == StorageValueType.text
            or value_type == StorageValueType.text_multiline
        ):
            lines = file.readlines()
            return (
                None
                if not lines or len(lines) == 0
                else (lines[0] if len(lines) == 1 else lines)
            )
        elif value_type == StorageValueType.dictionary:
            return YAML().load(file)


def delete_user_data(app: str, key: str):
    dir_path, file_name = _resolve_storage_path(app, key)
    full_path = os.path.join(dir_path, file_name)
    if os.path.isdir(full_path):
        rmtree(full_path)
    else:
        file_path, value_type = _storage_file(dir_path, file_name)
        if not value_type:
            return
        os.remove(file_path)


def clear_storage():
    storage_dir_path = _get_storage_dir_path()
    rmtree(storage_dir_path)
    os.mkdir(storage_dir_path)


def store_local_data(key: str, data: str):
    dir_path = get_local_data_dir()
    with open(os.path.join(dir_path, key), "a") as f:
        f.write(f"{data}\n")


def get_local_config_dir():
    return _get_storage_dir_path()


def get_local_data_dir():
    dir_path = __storage_path_by_os(StoragePurpose.data)[sys.platform]
    Path(dir_path).mkdir(exist_ok=True, parents=True)
    return dir_path



from typing import Optional

from pydantic import BaseModel


class Env(BaseModel):
    name: str
    alias: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None



import datetime

from hexagon.runtime.singletons import options
from hexagon.support.storage import (
    HexagonStorageKeys,
    load_user_data,
    store_user_data,
)

LAST_UPDATE_DATE_FORMAT = "%Y%m%d"


def already_checked_for_updates(app: str = None) -> bool:
    last_checked = load_user_data(HexagonStorageKeys.last_update_check.value, app)

    result = False

    if last_checked:
        last_checked_date = datetime.datetime.strptime(
            last_checked, LAST_UPDATE_DATE_FORMAT
        )

        result = (
            last_checked_date + options.update_time_between_checks
            >= datetime.datetime.now()
        )

    if not result:
        store_user_data(
            HexagonStorageKeys.last_update_check.value,
            datetime.date.today().strftime(LAST_UPDATE_DATE_FORMAT),
            app=app,
        )

    return result



from dataclasses import dataclass
from enum import Enum
from typing import List, Union, Optional

from hexagon.support.input.args import CliArgs


@dataclass
class Trace:
    ref: str
    value: str
    value_alias: str
    key: str
    key_alias: str

    def real(self):
        if self.key:
            return f"{self.key}={self.__safe_value}"
        return str(self.__safe_value)

    def alias(self):
        if self.key_alias:
            return f"{self.key_alias}={self.__safe_value}"
        if self.key:
            return f"{self.key}={self.__safe_value}"
        return str(self.value_alias or self.__safe_value)

    @property
    def __safe_value(self):
        """
        if value has spaces, wrap it in quotes to avoid parsing issues in the shell
        if value has single or double quotes, use the opposite to wrap it
        let's hope for the best
        """
        if " " in self.value:
            if '"' in self.value:
                return f"'{self.value}'"
            return f'"{self.value}"'
        return self.value


class Tracer:
    def __init__(self, initial_cli_args: CliArgs):
        self._initial_args = initial_cli_args
        self._trace: List[Trace] = []

    def tracing(
        self,
        ref: str,
        value: Union[str, Enum, list, bool],
        key: str = None,
        value_alias: str = None,
        key_alias: str = None,
    ):
        if not value:
            pass

        for t in self._trace:
            if t.ref == ref:
                self._trace.remove(t)

        def to_str(v):
            if isinstance(v, Enum):
                return str(v.value)
            elif isinstance(v, bool):
                return str(v).lower()
            return str(v)

        if not key:
            self._trace.append(Trace(ref, to_str(value), value_alias, key, key_alias))
            return value

        if isinstance(value, list):
            for val in value:
                self._trace.append(Trace(ref, to_str(val), value_alias, key, key_alias))
        else:
            self._trace.append(Trace(ref, to_str(value), value_alias, key, key_alias))
        return value

    def remove_last(self):
        del self._trace[-1]

    def trace(self):
        return " ".join([x.real() for x in self._trace])

    def aliases_trace(self):
        return " ".join([x.alias() for x in self._trace])

    def print_run_again(
        self,
        cli_command: str,
        logger,
    ):
        if self.has_traced():
            command, aliases_command = self.trace(), self.aliases_trace()
            logger.extra(
                _("msg.main.tracer.run_again").format(
                    command=f"{cli_command} {command}"
                )
            )
            if aliases_command and command != aliases_command:
                logger.extra(
                    _("msg.main.tracer.or").format(
                        command=f"{cli_command} {aliases_command}"
                    )
                )

    def has_traced(self):
        if len(self._trace) == 0:
            return False

        if len(self._trace) > self._initial_args.count():
            return True

        input_args = self._initial_args.as_str()
        matches = []
        for x in self._trace:
            if x.real() in input_args:
                matches.append(x.real())
            elif x.alias() in input_args:
                matches.append(x.alias())
            elif (
                x.real().replace("=", " ") in input_args
                or x.alias().replace("=", " ") in input_args
            ):
                matches.append(x.alias().replace("=", ""))

        return len(matches) > self._initial_args.count()


_tracer: Optional[Tracer] = None


def tracer() -> Tracer:
    global _tracer
    if not _tracer:
        raise ValueError("Tracer not initialized")
    return _tracer


def init_tracer(args: CliArgs):
    global _tracer
    _tracer = Tracer(args)
    return _tracer



import importlib
import os
import subprocess
import sys
from pathlib import Path
from typing import List, Union, Dict, Any, Optional

from pydantic import ValidationError

from hexagon.domain.env import Env
from hexagon.domain.hexagon_error import HexagonError
from hexagon.domain.tool import ActionTool
from hexagon.runtime.execute.errors import (
    ToolExecutionError,
    ActionInputError,
    ActionExecuteError,
    ActionImportError,
)
from hexagon.runtime.execute.execution_hook import execution_hook
from hexagon.runtime.parse_args import parse_cli_args
from hexagon.runtime.singletons import configuration
from hexagon.support.input.args import CliArgs, ToolArgs
from hexagon.support.input.prompt import Prompt
from hexagon.support.tracer import tracer

ENVVAR_EXECUTION_ENV = "HEXAGON_EXECUTION_ENV"

ENVVAR_EXECUTION_TOOL = "HEXAGON_EXECUTION_TOOL"

TOOL_ARGUMENTS_CLASS_NAME = "Args"

_command_by_file_extension = {"js": "node", "sh": "sh"}


@execution_hook
def execute_action(
    tool: ActionTool, env_args: Any, env: Optional[Env], cli_args: CliArgs
):
    custom_tools_path = configuration.custom_tools_path
    action_to_execute: str = tool.executable_str
    script_interpreter, script_abs_path = __get_interpreter_and_path(action_to_execute)

    if script_interpreter:
        _execute_script(
            script_interpreter,
            script_abs_path,
            env_args or [],
            tool,
            env,
            cli_args,
        )
    else:
        python_module_found = _execute_python_module(
            action_to_execute,
            tool,
            env,
            env_args,
            cli_args,
            custom_tools_path,
        )
        if python_module_found:
            return

        split_action = action_to_execute.split(" ")
        return_code, executed_command = _execute_inline_command(
            split_action[0],
            env_args,
            cli_args,
            tool,
            env,
            action_args=split_action[1:],
        )

        if return_code != 0:
            raise ToolExecutionError(
                return_code, executed_command, tool.executable_str, custom_tools_path
            )


def __get_interpreter_and_path(action_to_execute):
    # Script should be relative to the project path
    script_path = os.path.join(configuration.project_path, action_to_execute)
    f = Path(script_path)
    if not f.is_file():
        return None, None

    ext = f.suffix[1:]
    script_action_command = _command_by_file_extension.get(ext)
    return script_action_command, script_path


def _execute_python_module(
    action_id: str,
    tool: ActionTool,
    env: Env,
    env_args,
    cli_args,
    custom_tools_path: Optional[str],
):
    tool_action_module = _load_action_module(
        action_id, custom_tools_path
    ) or _load_action_module(f"hexagon.actions.external.{action_id}", custom_tools_path)

    if not tool_action_module:
        return False

    # noinspection PyBroadException
    try:
        tool_args = __parse_tool_args(cli_args, env, tool, tool_action_module)
        tool_action_module.main(
            tool,
            env,
            env_args,
            (
                tool_args
                if type(tool_args).__name__ == TOOL_ARGUMENTS_CLASS_NAME
                else tool_args.extra_args
            ),
        )
        return True
    except ValidationError as e:
        raise ActionInputError(e, tool.name)
    except HexagonError as e:
        raise e
    except Exception:
        raise ActionExecuteError(action_id, custom_tools_path)


def __parse_tool_args(cli_args, env, tool, tool_action_module):
    args = __args_with_optional_env(cli_args, env)
    # noinspection PyProtectedMember
    return (
        parse_cli_args(
            args,
            (
                tool_action_module.Args
                if hasattr(tool_action_module, TOOL_ARGUMENTS_CLASS_NAME)
                else ToolArgs
            ),
            prog=tool.name,
            description=tool.description or tool.long_name or "Hexagon tool",
            add_help=True,
            epilog=_("msg.support.execute.action.tool_help_epilog"),
        )
        ._with_tracer(tracer())
        ._with_prompt(Prompt())
    )


def __args_with_optional_env(cli_args: CliArgs, env: Optional[Env]):
    return (
        [cli_args.env.value] if not env and cli_args.env else []
    ) + cli_args.raw_extra_args


def _execute_inline_command(
    command: str,
    env_args: Any,
    cli_args: CliArgs,
    tool: ActionTool,
    env: Env = None,
    action_args: List[str] = None,
):
    action_args = action_args if action_args else []
    cmd_as_string = " ".join(
        [command] + action_args + __args_with_optional_env(cli_args, env)
    ).format(
        tool=tool,
        env=env,
        env_args=env_args,
        cli_args=cli_args.format_friendly_extra_args,  # TODO: make format_friendly_extra_args work here
    )

    return __call_subprocess(cmd_as_string, env, tool)


def _execute_script(
    command: str,
    script_path: str,
    env_args,
    tool: ActionTool,
    env: Env,
    cli_args: CliArgs,
):
    action_args = [script_path]
    hexagon_args = __sanitize_args_for_command(
        env_args, *__args_with_optional_env(cli_args, env)
    )
    cmd_as_string = " ".join([command] + action_args + hexagon_args).format(
        tool=tool,
        env=env,
        env_args=env_args,
        cli_args=cli_args.extra_args,
    )
    return __call_subprocess(cmd_as_string, env, tool)


def __sanitize_args_for_command(*args: Union[List[any], Dict, Env]):
    positional = []
    named = []
    for arg in args:
        if isinstance(arg, (int, float, complex, str)):
            positional.append(str(arg))
        elif isinstance(arg, list):
            for a in arg:
                positional.append(str(a))
        elif not arg:
            continue
        else:
            try:
                named += [f"{k}={v}" for k, v in arg.items() if v]
            except AttributeError:
                # Unknown arg type, try to append it directly
                positional.append(f'"{str(arg)}"')
    return named + positional


def _load_action_module(action_id: str, custom_tools_path):
    if not all(s and s.isidentifier() for s in action_id.split(".")):
        return None

    try:
        return __load_module(action_id)
    except ModuleNotFoundError as e:
        if e.name == action_id:
            return None
        else:
            raise ActionImportError(action_id, custom_tools_path)


def __load_module(module: str):
    if module in sys.modules:
        return sys.modules[module]

    return importlib.import_module(module)


def __call_subprocess(command: str, env: Optional[Env], tool: ActionTool):
    env_vars = os.environ.copy()
    env_vars[ENVVAR_EXECUTION_TOOL] = tool.json()
    if env:
        env_vars[ENVVAR_EXECUTION_ENV] = env.json()
    return (
        subprocess.call(command, shell=True, env=env_vars),
        command,
    )



import subprocess
import sys
from typing import Optional

from hexagon.runtime.dependencies.fs import declarations_found
from hexagon.support.output.printer import log

PYTHON_PIPFILE_NAME = "Pipfile"
PYTHON_REQUIREMENTS_NAME = "requirements.txt"
PYTHON_DEPENDENCY_FILES = [PYTHON_PIPFILE_NAME, PYTHON_REQUIREMENTS_NAME]


def scan_and_install_python_dependencies(path: str, mocked=False):
    for directory, files in declarations_found(path, PYTHON_DEPENDENCY_FILES):
        with log.status(
            _("msg.support.dependencies.installing_dependencies").format(
                runtime="python", path=directory
            )
        ):
            command: Optional[str] = None
            if PYTHON_REQUIREMENTS_NAME in files:
                command = f"python3 -m pip install -r {PYTHON_REQUIREMENTS_NAME}"
            elif PYTHON_PIPFILE_NAME in files:
                command = "pipenv install --system"

            if mocked:
                print(f"would have ran {command}")
            else:
                subprocess.check_call(
                    command,
                    shell=True,
                    cwd=directory,
                    stdout=sys.stdout,
                    stderr=subprocess.DEVNULL,
                )



import os

from pydantic import validator

from hexagon.support.input.args import ToolArgs, Arg, PositionalArg
from hexagon.support.output.printer import log
from hexagon.support.storage import (
    HexagonStorageKeys,
    load_user_data,
)


class Args(ToolArgs):
    alias_name: PositionalArg[str] = Arg(
        None,
        prompt_message=_("action.actions.internal.save_new_alias.prompt_alias_name"),
    )

    @validator("alias_name")
    def not_empty(cls, value):
        if not value:
            raise ValueError(
                _("error.actions.internal.save_new_alias.insert_valid_alias")
            )
        return value


def main(tool, env, env_args, cli_args: Args):
    last_command = load_user_data(HexagonStorageKeys.last_command.value)

    log.info(
        _("msg.actions.internal.save_new_alias.last_command").format(
            last_command=last_command
        )
    )

    cli_args.alias_name.prompt()

    save_new_alias(cli_args.alias_name.value, last_command)


def save_new_alias(alias_name, command):
    shell_ = (
        os.environ["HEXAGON_TEST_SHELL"]
        if "HEXAGON_TEST_SHELL" in os.environ
        else os.environ["SHELL"]
    )

    shell_aliases = {
        "/usr/bin/zsh": f'{os.environ["HOME"]}/.oh-my-zsh/custom/aliases.zsh',
        "/bin/zsh": f'{os.environ["HOME"]}/.oh-my-zsh/custom/aliases.zsh',
        "/usr/bin/bash": f'{os.environ["HOME"]}/.bash_aliases',
        "/bin/bash": f'{os.environ["HOME"]}/.bash_aliases',
        "HEXAGON_TEST_SHELL": "home-aliases.txt",
    }

    aliases_file_path = shell_aliases[shell_]

    with open(aliases_file_path, "a+") as aliases_file:
        aliases_file.write(
            "\n" "# added by hexagon\n" f'alias {alias_name}="{command}"'
        )
        aliases_file.seek(0)
        __pretty_print_created_alias(aliases_file, aliases_file_path, lines_to_show=-3)
        aliases_file.close()

    log.info(
        _("msg.actions.internal.save_new_alias.success"),
        gap_end=1,
    )
    log.result(f"[b]$ {alias_name}")
    log.info(
        _("msg.actions.internal.save_new_alias.execute_tip"),
        gap_start=1,
    )
    log.info(_("msg.actions.internal.save_new_alias.reload_builtins"))
    log.info(
        _("msg.actions.internal.save_new_alias.run_source").format(
            file_path=aliases_file_path
        )
    )


def __pretty_print_created_alias(aliases_file, file, lines_to_show=-10):
    log.gap()
    log.info(_("msg.actions.internal.save_new_alias.added_to_file").format(file=file))
    log.example("\n".join(aliases_file.read().splitlines()[lines_to_show:]))



import os

from hexagon.domain import CONFIG_FILE_ENV_VARIABLE_NAME
from hexagon.runtime.configuration import Configuration
from hexagon.runtime.options import get_options

configuration = Configuration()
cli, tools, envs = configuration.init_config(
    os.getenv(CONFIG_FILE_ENV_VARIABLE_NAME, "app.yaml")
)
options = get_options(cli.options or {})
